from dataclasses import dataclass, field
from typing import List, Optional, Tuple

from crunch_convert._model import RequirementLanguage
from crunch_convert.notebook._utils import list_of_string_factory


@dataclass()
class ImportedRequirement:
    alias: str
    name: Optional[str] = None
    extras: List[str] = field(default_factory=list_of_string_factory)
    specs: List[str] = field(default_factory=list_of_string_factory)
    language: RequirementLanguage = RequirementLanguage.PYTHON

    @property
    def extras_and_specs(self):
        return (self.extras, self.specs)

    def merge(self, other: "ImportedRequirement") -> Tuple[bool, Optional[str]]:
        """
        Merge requirements:
        - if name is missing or the same, then use other's name
        - if (extras and specs) are empty or the same, then use other's (extras and specs)

        Alias is ignored.
        """

        if self.language != other.language:
            raise ValueError(
                f"cannot merge requirements with different languages: "
                f"{self.language.name} != {other.language.name}"
            )

        errors: list[str] = []

        different_name = self.name is not None and other.name is not None and self.name != other.name
        if different_name:
            errors.append("name")

        different_extras = len(self.extras) and len(other.extras) and self.extras != other.extras
        if different_extras:
            errors.append("extras")

        different_specs = len(self.specs) and len(other.specs) and self.specs != other.specs
        if different_specs:
            errors.append("specs")

        if len(errors):
            error_count = len(errors)

            if error_count == 1:
                field = errors[0]
                be = "is" if field == "name" else "are"
                message = f"{field} {be} different"
            elif error_count == 2:
                message = f"both {errors[0]} and {errors[1]} are different"
            elif error_count == 3:
                message = f"{errors[0]}, {errors[1]} and {errors[2]} are all different"

            return False, message  # type: ignore

        if not different_name and other.name is not None:
            self.name = other.name

        if not different_extras and not different_specs and (len(other.extras) or len(other.specs)):
            self.extras = other.extras
            self.specs = other.specs

        return True, None


@dataclass()
class EmbeddedFile:
    path: str
    normalized_path: str
    content: str



from collections import defaultdict
from typing import DefaultDict, Dict, List, Optional, Union, cast

from crunch_convert._model import RequirementLanguage
from crunch_convert.notebook._model import ImportedRequirement
from crunch_convert.requirements_txt._model import NamedRequirement
from crunch_convert.requirements_txt._whitelist import Library, Whitelist

Requirement = Union[NamedRequirement, ImportedRequirement]


def format_files_from_imported(
    requirements: List[ImportedRequirement],
    *,
    header: str,
    whitelist: Whitelist,
) -> Dict[RequirementLanguage, str]:
    return _format_files(
        requirements=cast(List[Requirement], requirements),
        header=header,
        whitelist=whitelist,
    )


def format_files_from_named(
    requirements: List[NamedRequirement],
    *,
    header: str,
    whitelist: Whitelist,
) -> Dict[RequirementLanguage, str]:
    return _format_files(
        requirements=cast(List[Requirement], requirements),
        header=header,
        whitelist=whitelist,
    )


def _format_files(
    *,
    requirements: List[Requirement],
    header: str,
    whitelist: Whitelist,
) -> DefaultDict[RequirementLanguage, str]:
    files: DefaultDict[RequirementLanguage, str] = defaultdict(str)

    requirements_by_language: Dict[RequirementLanguage, List[Requirement]] = defaultdict(list)
    for requirement in requirements:
        requirements_by_language[requirement.language].append(requirement)

    for language, requirements_for_language in requirements_by_language.items():
        files[language] = _format_files_filtered(
            requirements_for_language,
            header=header,
            whitelist=whitelist,
        )

    return files


def _format_files_filtered(
    requirements: List[Requirement],
    *,
    header: str,
    whitelist: Whitelist,
):
    third_party: List[str] = []
    standard: List[str] = []

    for requirement in requirements:
        library = find_library(requirement, whitelist)

        if library is None:
            third_party.append(format_line(
                get_fallback_name(requirement),
                None,
                requirement.extras,
                requirement.specs,
            ))

        elif library.standard:
            standard.append(get_fallback_name(requirement))

        elif isinstance(requirement, ImportedRequirement):
            third_party.append(format_line(
                library.name,
                requirement.alias,
                requirement.extras,
                requirement.specs,
            ))

        else:
            third_party.append(format_line(
                library.name,
                None,
                requirement.extras,
                requirement.specs,
            ))

    third_party.sort(key=str.lower)
    standard.sort(key=str.lower)

    builder = f"# {header}\n"

    if third_party:
        builder += "\n"

        if standard:
            builder += "## third-party\n"

        for requirement_line in third_party:
            builder += f"{requirement_line}\n"

    if standard:
        builder += "\n"

        if third_party:
            builder += "## standard\n"

        for requirement_line in standard:
            builder += f"#{requirement_line}\n"

    return builder


def format_line(
    name: str,
    alias: Optional[str],
    extras: List[str],
    specs: List[str],
) -> str:
    line = name

    if extras:
        line += f"[{','.join(extras)}]"

    if specs:
        line += f"{','.join(specs)}"

    if alias and name != alias:
        line += f"  # alias of {alias}"

    return line


def get_fallback_name(
    requirement: Requirement,
) -> str:
    if isinstance(requirement, ImportedRequirement):
        return requirement.name or requirement.alias

    return requirement.name


def find_library(
    requirement: Requirement,
    whitelist: Whitelist,
) -> Optional[Library]:
    if isinstance(requirement, ImportedRequirement):
        name = requirement.name
        if name:
            return whitelist.find_library(
                language=requirement.language,
                name=name,
            )

        return whitelist.find_library(
            language=requirement.language,
            alias=requirement.alias,
        )
    else:
        assert isinstance(requirement, NamedRequirement)

        return whitelist.find_library(
            language=requirement.language,
            name=requirement.name,
        )



from crunch_convert._model import RequirementLanguage as RequirementLanguage
from crunch_convert.requirements_txt._formatter import \
    format_files_from_imported as format_files_from_imported
from crunch_convert.requirements_txt._formatter import \
    format_files_from_named as format_files_from_named
from crunch_convert.requirements_txt._model import \
    NamedRequirement as NamedRequirement
from crunch_convert.requirements_txt._whitelist import \
    CachedWhitelist as CachedWhitelist
from crunch_convert.requirements_txt._whitelist import \
    CrunchHubWhitelist as CrunchHubWhitelist
from crunch_convert.requirements_txt._whitelist import Library as Library
from crunch_convert.requirements_txt._whitelist import \
    LocalWhitelist as LocalWhitelist
from crunch_convert.requirements_txt._whitelist import Whitelist as Whitelist



import ast
import json
import os
import py_compile
import re
from collections import defaultdict
from dataclasses import dataclass
from typing import (Any, Callable, DefaultDict, Dict, List, Match, NamedTuple,
                    Optional, Set, Tuple, Union, cast)

import libcst
import yaml

import requirements
from crunch_convert._model import RequirementLanguage
from crunch_convert.notebook._model import (EmbeddedFile, ImportedRequirement)
from crunch_convert.notebook._r import is_r_import
from crunch_convert.notebook._utils import cut_crlf, strip_hashes

_FAKE_PACKAGE_NAME = "x__fake_package_name__"
_PACKAGE_NAME_PATTERN = r"[a-zA-Z_][a-zA-Z0-9_-]*[a-zA-Z0-9]"
_LAST_VERSION = "@latest"

_DOT = "."
_KV_DIVIDER = "---"

_CRUNCH_KEEP_ON = "@crunch/keep:on"
_CRUNCH_KEEP_OFF = "@crunch/keep:off"

JUPYTER_MAGIC_COMMAND_PATTERN = r"^(\s*?)(!|%|pip3? )"

_EMPTY_EXTRAS_AND_SPECS: Tuple[List[str], List[str]] = ([], [])


LogFunction = Callable[[str], None]


def strip_packages(name: str):
    if name.startswith(_DOT):
        return None  # just in case, but should not happen

    if _DOT not in name:
        return name

    index = name.index(_DOT)
    return name[:index]


class ConverterError(ValueError):
    pass


class NotebookCellParseError(ConverterError):

    def __init__(
        self,
        message: str,
        parser_error: Optional[str],
        cell_source: str,
        cell_index: Optional[int] = None,
        cell_id: Optional[str] = None,
    ) -> None:
        super().__init__(message)
        self.parser_error = parser_error
        self.cell_index = cell_index
        self.cell_id = cell_id
        self.cell_source = cell_source


class RequirementVersionParseError(ConverterError):

    def __init__(self, message: str) -> None:
        super().__init__(message)


class ImportInfo(NamedTuple):
    name: Optional[str]
    extras: List[str]
    specs: List[str]


class InconsistantLibraryVersionError(ConverterError):

    def __init__(
        self,
        message: str,
        package_name: str,
        old: ImportInfo,
        new: ImportInfo,
    ) -> None:
        super().__init__(message)
        self.package_name = package_name
        self.old = old
        self.new = new


def _extract_import_version(
    log: LogFunction,
    comment_node: Optional[libcst.Comment]
) -> Optional[ImportInfo]:
    if comment_node is None:
        log(f"skip version: no comment")
        return None

    line = strip_hashes(comment_node.value)
    if not line:
        log(f"skip version: comment empty")
        return None

    match = re.match(r"^(" + _PACKAGE_NAME_PATTERN + r")?\s*([@\[><=~])", line)
    if not match:
        log(f"skip version: line not matching: `{line}`")
        return None

    user_package_name = match.group(1)
    test_package_name = user_package_name or _FAKE_PACKAGE_NAME

    version_part: str = line[match.start(2):]
    if version_part == _LAST_VERSION:
        return ImportInfo(user_package_name or None, [], [])

    line = f"{test_package_name} {version_part}"

    try:
        requirement = next(requirements.parse(line), None)
        if requirement is None:
            log(f"skip version: parse returned nothing: `{line}`")
            return None
    except Exception as error:
        raise RequirementVersionParseError(
            f"version cannot be parsed: {error}"
        ) from error

    if requirement.name != test_package_name:
        # package has been modified somehow
        raise RequirementVersionParseError(
            f"name must be `{test_package_name}` and not `{requirement.name}`"
        )

    return ImportInfo(
        user_package_name or None,
        list(requirement.extras),
        [
            f"{operator}{semver}"
            for operator, semver in requirement.specs
        ]
    )


ImportNodeType = Union[libcst.Import, libcst.ImportFrom]


def _evaluate_name(node: libcst.CSTNode) -> str:
    if isinstance(node, libcst.Name):
        return node.value
    elif isinstance(node, libcst.Attribute):
        return f"{_evaluate_name(node.value)}.{node.attr.value}"
    else:
        raise Exception("Logic error!")


def _convert_python_import(
    log: LogFunction,
    import_node: ImportNodeType,
    comment_node: Optional[libcst.Comment]
) -> List[ImportedRequirement]:
    if isinstance(import_node, libcst.Import):
        paths = [
            _evaluate_name(alias.name)
            for alias in import_node.names
        ]
    elif isinstance(import_node, libcst.ImportFrom) and import_node.module is not None:  # type: ignore
        paths = [_evaluate_name(import_node.module)]
    else:
        return []

    (
        package_name,
        extras,
        specs,
    ) = _extract_import_version(
        log,
        comment_node
    ) or (None, [], [])

    names: Set[str] = set()
    for path in paths:
        name = strip_packages(path)
        if name:
            names.add(name)

    return [
        ImportedRequirement(
            alias=name,
            name=package_name,
            extras=extras,
            specs=specs,
            language=RequirementLanguage.PYTHON
        )
        for name in names
    ]


def _add_to_packages(
    imported_requirements: Dict[str, ImportedRequirement],
    new_requirements: List[ImportedRequirement],
    log: LogFunction,
):
    for new in new_requirements:
        package_name = new.alias

        if package_name in imported_requirements:
            current = imported_requirements[package_name]

            success, message = current.merge(new)
            if not success:
                raise InconsistantLibraryVersionError(
                    f"inconsistant requirements for the same package: {message}",
                    package_name,
                    ImportInfo(current.name, current.extras, current.specs),
                    ImportInfo(new.name, new.extras, new.specs),
                )
        else:
            imported_requirements[package_name] = new

        if new.extras_and_specs != _EMPTY_EXTRAS_AND_SPECS:
            log(f"found version: {package_name}: {new}")


_IMPORT = (
    libcst.Import,
    libcst.ImportFrom,
)

_KEEP = (
    libcst.Module,

    libcst.FunctionDef,
    libcst.ClassDef,

    libcst.Comment,
    libcst.EmptyLine,
    libcst.TrailingWhitespace,

    libcst.SimpleStatementLine,
)


class Comment(libcst.Comment):

    semicolon = False

    def _codegen_impl(self, state, default_semicolon=None) -> None:  # type: ignore
        super()._codegen_impl(state)  # type: ignore


class EmptyLine(libcst.EmptyLine):

    semicolon = False

    def _codegen_impl(self, state, default_semicolon=None) -> None:  # type: ignore
        super()._codegen_impl(state)  # type: ignore


class CommentTransformer(libcst.CSTTransformer):

    METHOD_GROUP = "group"
    METHOD_LINE = "line"

    def __init__(self, tree: libcst.Module):
        self.tree = tree

        self.import_and_comment_nodes: List[Tuple[ImportNodeType, Optional[libcst.Comment]]] = []
        self.r_import_names: List[str] = []

        self._method_stack: List[str] = []
        self._previous_import_node: Optional[ImportNodeType] = None
        self._auto_comment = True

    def on_visit(
        self,
        node: libcst.CSTNode,
    ) -> bool:
        # print("visit", type(node), "\\n".join(self._to_lines(node)))

        if isinstance(node, (libcst.Module, libcst.SimpleStatementLine)):
            self._method_stack.append(self.METHOD_GROUP)
            return True
        elif isinstance(node, libcst.BaseCompoundStatement):
            self._method_stack.append(self.METHOD_GROUP)
            return False
        else:
            self._method_stack.append(self.METHOD_LINE)
            return False

    def on_leave(
        self,
        original_node: libcst.CSTNode,
        updated_node: libcst.CSTNode
    ) -> Union[libcst.CSTNode, libcst.RemovalSentinel, libcst.FlattenSentinel[libcst.CSTNode]]:
        method = self._method_stack.pop()
        # print("leave", type(original_node), method)

        if isinstance(original_node, _IMPORT):
            self._previous_import_node = original_node
            return updated_node

        r_requirement = is_r_import(original_node)
        if r_requirement is not None:
            self.r_import_names.append(r_requirement)
            self._previous_import_node = cast(ImportNodeType, original_node)
            return updated_node

        if self._previous_import_node is not None:
            import_node, self._previous_import_node = self._previous_import_node, None

            if isinstance(original_node, libcst.TrailingWhitespace) and original_node.comment:
                self.import_and_comment_nodes.append(
                    (import_node, original_node.comment)
                )
            else:
                self.import_and_comment_nodes.append(
                    (import_node, None)
                )

        if isinstance(original_node, libcst.EmptyLine) and original_node.comment:
            comment = strip_hashes(original_node.comment.value)
            if comment == _CRUNCH_KEEP_ON:
                self._auto_comment = False
            elif comment == _CRUNCH_KEEP_OFF:
                self._auto_comment = True

            return updated_node

        if not self._auto_comment or isinstance(original_node, _KEEP):
            return updated_node

        nodes: List[libcst.CSTNode] = []

        # control flow blocks have their comment attached to them
        if isinstance(original_node, libcst.BaseCompoundStatement) and original_node.leading_lines:
            nodes.extend(original_node.leading_lines)

            original_node = original_node.with_changes(
                leading_lines=libcst.FlattenSentinel([])
            )

        if method == self.METHOD_GROUP:
            nodes.extend(
                EmptyLine(comment=Comment(f"#{line}"))
                for line in self._to_lines(original_node)
            )

        elif method == self.METHOD_LINE:
            if isinstance(original_node, libcst.BaseSmallStatement):
                lines = self._to_lines(original_node)

                if len(lines) == 1:
                    nodes.append(Comment(f"#{lines[0]}"))
                else:
                    nodes.extend(
                        EmptyLine(comment=Comment(f"#{line}"))
                        for line in lines[:-1]
                    )
                    nodes.append(Comment(f"#{lines[-1]}"))
            else:
                nodes.extend(
                    Comment(f"#{line}")
                    for line in self._to_lines(original_node)
                )

        else:
            raise NotImplementedError(f"method: {method}")

        return libcst.FlattenSentinel(nodes)

    def _to_lines(self, node: libcst.CSTNode) -> List[str]:
        return self.tree.code_for_node(node).splitlines()


def _jupyter_replacer(match: Match[str]) -> str:
    spaces = match.group(1)
    command = match.group(2)

    if len(spaces):
        return f"{spaces}pass  #{command}"

    return f"#{command}"


def _extract_code_cell(
    cell_source: List[str],
    log: LogFunction,
    module: List[str],
    imported_requirements: DefaultDict[RequirementLanguage, Dict[str, ImportedRequirement]],
):
    source = "\n".join(
        re.sub(JUPYTER_MAGIC_COMMAND_PATTERN, _jupyter_replacer, line)
        for line in cell_source
    )

    if not len(source):
        log(f"skip since empty")
        return

    try:
        tree = libcst.parse_module(source)
    except libcst.ParserSyntaxError as error:
        log(f"failed to parse: {error.message}")

        raise NotebookCellParseError(
            f"notebook code cell cannot be parsed",
            str(error),
            source,
        ) from error

    transformer = CommentTransformer(tree)
    tree = tree.visit(transformer)

    for import_node, comment_node in transformer.import_and_comment_nodes:
        new_requirements = _convert_python_import(log, import_node, comment_node)

        _add_to_packages(
            imported_requirements[RequirementLanguage.PYTHON],
            new_requirements,
            log,
        )

    if True:
        new_requirements = [
            ImportedRequirement(
                alias=name,
                language=RequirementLanguage.R
            )
            for name in transformer.r_import_names
        ]

        _add_to_packages(
            imported_requirements[RequirementLanguage.R],
            new_requirements,
            log,
        )

    lines = tree.code.strip("\r\n").splitlines()
    if len(lines):
        log(f"used {len(lines)} line(s)")

        if len(module):
            module.append(f"\n")

        module.append("\n".join(lines))
    else:
        log(f"skip since empty")


def _extract_markdown_cell(
    cell_source: List[str],
    log: LogFunction,
    embedded_files: Dict[str, EmbeddedFile],
):
    if not len(cell_source):
        log(f"skip since empty")
        return

    def get_full_source():
        return "\n".join(cell_source)

    iterator = iter(cell_source)

    if next(iterator) != _KV_DIVIDER:
        return

    source: Union[List[str], str] = []
    valid = True

    try:
        for line in iterator:
            if not line.strip():
                valid = False
                break

            if line == _KV_DIVIDER:
                break

            source.append(line)
        else:
            valid = False

        if not valid:
            return

        source = "\n".join(source)

        configuration: Dict[str, Any] = yaml.safe_load(source)

        if not isinstance(configuration, dict):  # type: ignore
            raise ValueError("root must be a dict")
    except Exception as error:
        log(f"failed to parse: {error}")

        if isinstance(source, list):
            source = "\n".join(source)

        raise NotebookCellParseError(
            f"notebook markdown cell cannot be parsed",
            str(error),
            source,
        ) from error

    file_path = configuration.get("file")
    if not isinstance(file_path, str) or not file_path.strip():
        raise NotebookCellParseError(
            f"file not specified",
            None,
            get_full_source(),
        )

    normalized_file_path = os.path.normpath(file_path).replace("\\", "/")
    lower_file_path = normalized_file_path.lower()

    previous = embedded_files.get(lower_file_path)
    if previous is not None:
        raise NotebookCellParseError(
            f"file `{file_path}` specified multiple time",
            f"file `{file_path}` is conflicting with `{previous.path}`",
            get_full_source(),
        )

    content = "\n".join(iterator).strip()

    embedded_files[lower_file_path] = EmbeddedFile(
        path=file_path,
        normalized_path=normalized_file_path,
        content=content,
    )

    log(f"embed {lower_file_path}: {len(content)} characters")


def _validate(source_code: str):
    try:
        ast.parse(source_code)
    except SyntaxError as error:
        parser_error = py_compile.PyCompileError(
            error.__class__,
            error,
            "<converted_output>",
        )

        raise NotebookCellParseError(
            f"converted notebook code cell cannot be compiled",
            str(parser_error),
            source_code,
            -1,
            parser_error.file
        )


@dataclass(frozen=True)
class FlattenNotebook:
    source_code: str
    embedded_files: List[EmbeddedFile]
    requirements: List["ImportedRequirement"]


def extract_from_cells(
    cells: List[Any],
    *,
    print: Optional[LogFunction] = print,
    validate: bool = True,
) -> FlattenNotebook:
    if print is None:
        def print(_): return None

    imported_requirements: DefaultDict[RequirementLanguage, Dict[str, ImportedRequirement]] = defaultdict(dict)
    module: List[str] = []
    embed_files: Dict[str, EmbeddedFile] = {}

    for index, cell in enumerate(cells):
        cell_id = cell["metadata"].get("id") or f"cell_{index}"

        def log(message: str):
            print(f"convert {cell_id}: {message}")

        cell_source = cell["source"]
        if isinstance(cell_source, str):
            cell_source = cell_source.split("\n")

        cell_source = [
            cut_crlf(line)
            for line in cell_source
        ]

        try:
            cell_type = cell["cell_type"]
            if cell_type == "code":
                _extract_code_cell(cell_source, log, module, imported_requirements)
            elif cell_type == "markdown":
                _extract_markdown_cell(cell_source, log, embed_files)
            else:
                log(f"skip since unknown type: {cell_type}")
                continue
        except NotebookCellParseError as error:
            error.cell_index = index
            error.cell_id = cell_id
            raise

    module.append("")
    source_code = "\n".join(module)

    if validate:
        _validate(source_code)

    return FlattenNotebook(
        source_code=source_code,
        embedded_files=list(embed_files.values()),
        requirements=[
            requirement
            for requirements in imported_requirements.values()
            for requirement in list(requirements.values())
        ],
    )


def extract_from_file(
    path: str,
    *,
    print: Optional[LogFunction] = None,
    validate: bool = True,
) -> FlattenNotebook:
    with open(path) as fd:
        notebook = json.load(fd)

    cells = notebook["cells"]

    return extract_from_cells(
        cells,
        print=print,
        validate=validate,
    )



"""
crunch-convert
~~~~~~
The crunch module to convert a notebook into multiple components!
"""

from crunch_convert import notebook as notebook
from crunch_convert import requirements_txt as requirements_txt
from crunch_convert._model import RequirementLanguage as RequirementLanguage



from enum import Enum
from typing import Any, cast


class RequirementLanguage(Enum):
    PYTHON = "PYTHON", "requirements.txt"
    R = "R", "requirements.r.txt"

    # https://stackoverflow.com/questions/12680080/python-enums-with-attributes/19300424#19300424
    def __new__(
        cls,
        *args: Any,
        **kwds: Any,
    ) -> "RequirementLanguage":
        obj = object.__new__(cls)
        obj._value_ = cast(str, args[0])
        return obj

    def __init__(
        self,
        value: str,
        txt_file_name: str,
    ):
        value; # type: ignore
        self.txt_file_name = txt_file_name



from string import whitespace
from textwrap import indent
from typing import List


def list_of_string_factory() -> List[str]:
    return []


def print_indented(text: str):
    indented = indent(text, "   | ", lambda x: True)

    if indented.endswith("\n"):
        indented = indented[:-1]

    print(indented)


def cut_crlf(input: str):
    input = input.replace("\r", "")

    if input.endswith('\n'):
        input = input[:-1]

    return input


def strip_hashes(input: str) -> str:
    return input.strip(whitespace + "#")



from crunch_convert import RequirementLanguage


def test_r():
    assert "R" == RequirementLanguage.R.value
    assert "requirements.r.txt" == RequirementLanguage.R.txt_file_name


def test_python():
    assert "PYTHON" == RequirementLanguage.PYTHON.value
    assert "requirements.txt" == RequirementLanguage.PYTHON.txt_file_name



from crunch_convert._model import RequirementLanguage as RequirementLanguage
from crunch_convert.notebook._model import EmbeddedFile as EmbeddedFile
from crunch_convert.notebook._model import \
    ImportedRequirement as ImportedRequirement
from crunch_convert.notebook._notebook import ConverterError as ConverterError
from crunch_convert.notebook._notebook import \
    FlattenNotebook as FlattenNotebook
from crunch_convert.notebook._notebook import \
    InconsistantLibraryVersionError as InconsistantLibraryVersionError
from crunch_convert.notebook._notebook import \
    NotebookCellParseError as NotebookCellParseError
from crunch_convert.notebook._notebook import \
    RequirementVersionParseError as RequirementVersionParseError
from crunch_convert.notebook._notebook import \
    extract_from_cells as extract_from_cells
from crunch_convert.notebook._notebook import \
    extract_from_file as extract_from_file

# alias for compatibility with previous versions
EmbedFile = EmbeddedFile
ImportedRequirementLanguage = RequirementLanguage



import json
import os

import click

from crunch_convert.__version__ import __version__
from crunch_convert.notebook import (ConverterError,
                                     InconsistantLibraryVersionError,
                                     NotebookCellParseError, extract_from_file)
from crunch_convert.notebook._utils import print_indented


@click.group()
@click.version_option(__version__, package_name="__version__.__title__")
def cli():
    pass


@cli.command(help="Convert a notebook to a python script.")
@click.option("--override", is_flag=True, help="Force overwrite of the python file.")
@click.argument("notebook-file-path", required=True)
@click.argument("python-file-path", default="main.py")
def notebook(
    override: bool,
    notebook_file_path: str,
    python_file_path: str,
):
    try:
        flatten = extract_from_file(
            notebook_file_path,
            print=print,
            validate=True,
        )
    except IOError as error:
        print(f"{notebook_file_path}: cannot read notebook file: {error}")
        raise click.Abort()
    except json.JSONDecodeError as error:
        print(f"{notebook_file_path}: cannot parse notebook file: {error}")
        raise click.Abort()
    except ConverterError as error:
        print(f"{notebook_file_path}: convert failed: {error}")

        if isinstance(error, NotebookCellParseError):
            print(f"  cell: {error.cell_id} ({error.cell_index})")
            print(f"  source:")
            print_indented(error.cell_source)
            print(f"  parser error:")
            print_indented(error.parser_error or "None")

        elif isinstance(error, InconsistantLibraryVersionError):
            print(f"  package name: {error.package_name}")
            print(f"  first version: {error.old}")
            print(f"  other version: {error.new}")

        raise click.Abort()

    if not override and os.path.exists(python_file_path):
        override = click.prompt(
            f"file {python_file_path} already exists, override?",
            type=bool,
            default=False,
            prompt_suffix=" "
        )

        if not override:
            raise click.Abort()

    with open(python_file_path, "w") as fd:
        fd.write(flatten.source_code)



from abc import ABC, abstractmethod
from collections import defaultdict
from dataclasses import dataclass
from io import UnsupportedOperation
from typing import Any, Dict, List, Optional, Set, Tuple, cast, overload

import requests

from crunch_convert._model import RequirementLanguage


class MultipleLibraryAliasCandidateException(Exception):

    def __init__(self, alias: str, names: Set[str]):
        super().__init__(f"multiple library match the alias")
        self.alias = alias
        self.names = names


@dataclass()
class Library:
    language: RequirementLanguage
    name: str
    alias: Optional[str]
    standard: bool


class Whitelist(ABC):

    @overload
    def find_library(
        self,
        *,
        language: RequirementLanguage = RequirementLanguage.PYTHON,
        name: str,
    ) -> Optional[Library]:
        ...

    @overload
    def find_library(
        self,
        *,
        language: RequirementLanguage = RequirementLanguage.PYTHON,
        alias: str,
    ) -> Optional[Library]:
        ...

    @abstractmethod
    def find_library(
        self,
        *,
        language: RequirementLanguage = RequirementLanguage.PYTHON,
        name: Optional[str] = None,
        alias: Optional[str] = None,
    ) -> Optional[Library]:
        pass


CacheCompositeKey = Tuple[RequirementLanguage, str]


class CachedWhitelist(Whitelist):

    def __init__(
        self,
        delegate: Whitelist,
    ):
        self._delegate = delegate

        self._name_cache: dict[CacheCompositeKey, Library] = {}
        self._alias_cache: dict[CacheCompositeKey, Library] = {}

    def find_library(
        self,
        *,
        language: RequirementLanguage = RequirementLanguage.PYTHON,
        name: Optional[str] = None,
        alias: Optional[str] = None,
    ) -> Optional[Library]:
        if name is not None:
            cached = self._name_cache.get((language, name))
            if cached is not None:
                return cached

            library = self._delegate.find_library(
                language=language,
                name=name,
            )

        elif alias is not None:
            cached = self._alias_cache.get((language, alias))
            if cached is not None:
                return cached

            library = self._delegate.find_library(
                language=language,
                alias=alias,
            )

        else:
            raise UnsupportedOperation("find_library() must be called with either 'name' or 'alias'.")

        self._store_in_cache(library)
        return library

    def _store_in_cache(
        self,
        library: Optional[Library],
    ) -> None:
        if library is None:
            return

        self._name_cache[(library.language, library.name)] = library

        alias = library.alias
        if alias is not None:
            self._alias_cache[(library.language, alias)] = library


class CrunchHubWhitelist(Whitelist):

    def __init__(
        self,
        api_base_url: str = "https://api.hub.crunchdao.com/",
    ):
        self._api_base_url = api_base_url

    def find_library(
        self,
        *,
        language: RequirementLanguage = RequirementLanguage.PYTHON,
        name: Optional[str] = None,
        alias: Optional[str] = None,
    ) -> Optional[Library]:
        response = requests.get(
            f"{self._api_base_url}/v2/libraries",
            params={
                "language": language.value,
                # TODO Use dedicated parameter for alias
                "name": name or alias
            },
        )

        if response.status_code != 200:
            return None

        content = response.json().get("content", [])
        if not content:
            return None

        alias_conflicting_names: Set[str] = set()
        library: Optional[Library] = None

        for item in content:
            library = self._map_library(item)

            if name is not None and name == library.name:
                return library

            elif alias is not None and alias == library.alias:
                alias_conflicting_names.add(library.name)

            else:
                library = None

        if len(alias_conflicting_names) > 1:
            raise MultipleLibraryAliasCandidateException(
                alias=cast(str, alias),
                names=alias_conflicting_names,
            )

        return library

    def _map_library(
        self,
        item: Dict[str, Any],
    ) -> Library:
        aliases = item.get("aliases")
        alias = aliases[0] if aliases else None

        return Library(
            language=RequirementLanguage(item["language"]),
            name=item["name"],
            alias=alias,
            standard=item["standard"],
        )


class LocalWhitelist(Whitelist):

    def __init__(
        self,
        libraries: List[Library]
    ):
        self._library_by_name: Dict[CacheCompositeKey, Library] = {}
        self._libraries_by_alias: Dict[CacheCompositeKey, List[Library]] = defaultdict(list)

        for library in libraries:
            self._library_by_name[(library.language, library.name)] = library

            if library.alias is not None:
                self._libraries_by_alias[(library.language, library.alias)].append(library)

    def find_library(
        self,
        *,
        language: RequirementLanguage = RequirementLanguage.PYTHON,
        name: Optional[str] = None,
        alias: Optional[str] = None,
    ) -> Optional[Library]:
        if name is not None:
            return self._library_by_name.get((language, name))

        elif alias is not None:
            libraries = self._libraries_by_alias.get((language, alias))
            if not libraries:
                return None

            if len(libraries) > 1:
                names = set(
                    library.name
                    for library in libraries
                )

                raise MultipleLibraryAliasCandidateException(
                    alias=alias,
                    names=names,
                )

            return libraries[0]

        else:
            raise UnsupportedOperation("find_library() must be called with either 'name' or 'alias'.")



#!/usr/bin/env python3

import distutils.sysconfig
import os
from typing import Any

from setuptools import find_packages, setup

package = "crunch_convert"


about: dict[str, Any] = {}
here = os.path.abspath(os.path.dirname(__file__))
with open(os.path.join(here, package, '__version__.py')) as f:
    exec(f.read(), about)

with open('requirements/default.txt') as fd:
    requirements = fd.read().splitlines()

with open('requirements/test.txt') as fd:
    test_requirements = fd.read().splitlines()

with open('README.md') as fd:
    readme = fd.read()

setup(
    name=about['__title__'],
    description=about['__description__'],
    long_description=readme,
    long_description_content_type='text/markdown',
    version=about['__version__'],
    author=about['__author__'],
    author_email=about['__author_email__'],
    url=about['__url__'],
    packages=find_packages(),
    python_requires=">=3",
    install_requires=requirements,
    extras_require={
        'test': test_requirements,
    },
    zip_safe=False,
    entry_points={
        'console_scripts': ['crunch-convert=crunch_convert.cli:cli'],
    },
    classifiers=[
        'Intended Audience :: Developers',
        'Programming Language :: Python :: 3.7',
    ],
    keywords='package development template'
)
